/*!
 * Copyright (C) 2005 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * \file turtleboot.c
 * \brief ATmega8 boot loader.
 *
 * This is a boot loader for the Turtelizer programming adapter
 * and uses the Turtelizer protocol to update the application
 * section of an ATmega8.
 *
 * \verbatim
 *
 * $Log: turtleboot.c,v $
 * Revision 1.1  2006/01/09 09:17:40  haraldkipp
 * First check-in.
 *
 *
 * \endverbatim
 */

#include <avr/io.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>

#include "comm.h"

/*! \brief Turtelizer command to start the firmware update. */
#define FIRMWARE_START  '0'
/*! \brief Turtelizer command to erase a firmware page. */
#define FIRMWARE_ERASE  '1'
/*! \brief Turtelizer command to write a firmware pagge. */
#define FIRMWARE_WRITE  '2'
/*! \brief Turtelizer command to read firmware page. */
#define FIRMWARE_READ   '3'
/*! \brief Turtelizer command to exit the firmware update. */
#define FIRMWARE_END    '4'

static unsigned short addr;

/*!
 * \brief Process a firmware read telegram.
 *
 * - Request: FIRMWARE_START(1)
 * - Response: RESULT(1) VERSION(4)
 */
void DoFirmwareStart(void)
{
    tbuf[0] = 0;
    tbuf[1] = (unsigned char)VERSION;
    tbuf[2] = (unsigned char)(VERSION >> 8);
    tbuf[3] = (unsigned char)(VERSION >> 16);
    tbuf[4] = (unsigned char)(VERSION >> 24);
    tlen = 5;
}

/*!
 * \brief Process a firmware read telegram.
 *
 * - Request: FIRMWARE_READ(1) ADDRESS(4)
 * - Response: RESULT(1) PAGESIZE(4) DATA(n)
 */
void DoFirmwareRead(void)
{
    unsigned short i;

    tbuf[0] = 0;
    tbuf[1] = (unsigned char) SPM_PAGESIZE;
    tbuf[2] = (unsigned char) (SPM_PAGESIZE >> 8);
    tbuf[3] = 0;
    tbuf[4] = 0;
    for (i = 0; i < SPM_PAGESIZE; i++) {
        tbuf[i + 5] = pgm_read_byte_near(addr + i);
    }
    tlen = SPM_PAGESIZE + 5;
}

/*!
 * \brief Process a firmware erase telegram.
 *
 * - Request: FIRMWARE_ERASE(1) ADDRESS(4)
 * - Response: RESULT(1)
 */
void DoFirmwareErase(void)
{
    boot_page_erase(addr);
    boot_spm_busy_wait();

    tbuf[0] = 0;
    tlen = 1;
}

/*!
 * \brief Process a firmware write telegram.
 *
 * - Request: FIRMWARE_WRITE(1) ADDRESS(4) DATA(n)
 * - Response: RESULT(1)
 */
void DoFirmwareWrite(void)
{
    unsigned short data;
    unsigned short i;
    unsigned char *tp = tbuf + 5;

    for (i = 0; i < tlen - 5; i += 2) {
        data = *tp++;
        data += (unsigned short) (*tp++) << 8;
        boot_page_fill(addr + i, data);
    }
    boot_page_write(addr);
    boot_spm_busy_wait();
    boot_rww_enable();

    tbuf[0] = 0;
    tlen = 1;
}

int main(void)
{
    unsigned char done = 0;

    /* Enable UART communication. */
    InitUart();

    /* 
     * Initially send a positive response telegram. This
     * is done because the Turtelizer application doesn't
     * answer the telegram that activates the boot loader.
     */
    DoFirmwareStart();
    for (;;) {
        /* Send our response. */
        SendTelegram();

        /* If previous command was FIRMWARE_END. */
        if (done) {
            break;
        }

        /* Wait for the next command telegram. */
        RecvTelegram();

        /* Most telegrams contain an address. */
        addr = (unsigned short) tbuf[1] + ((unsigned short) tbuf[2] << 8);

        switch (tbuf[0]) {
        case FIRMWARE_START:
            DoFirmwareStart();
            break;
        case FIRMWARE_READ:
            DoFirmwareRead();
            break;
        case FIRMWARE_ERASE:
            DoFirmwareErase();
            break;
        case FIRMWARE_WRITE:
            DoFirmwareWrite();
            break;
        case FIRMWARE_END:
            done = 1;
            tbuf[0] = 0;
            tlen = 1;
            break;
        default:
            tbuf[0] = 16;
            tlen = 1;
            break;
        }
    }
    return 0;
}
